Package: fib
Type: Package
Title: Package computing the fibonnaci sequences
Version: 1.0.0
Date: 2024-05-28
Author: Lennart Tuijnder
Maintainer: Lennart Tuijnder <lennart.tuijnder@openanalytics.eu>
Description: 'fib' is a package that compute the fibonnaci sequence however it contains a bug!
   This is actually on purpose as this package was crated for demonstration purposes for the 
   tinytest2JUnit package!
License: file LICENSE
RoxygenNote: 7.2.3
Encoding: UTF-8
Suggests: tinytest

Package: tinytest2JUnit
Type: Package
Title: Convert 'tinytest' Output to JUnit XML
Version: 1.1.3
Maintainer: Lennart Tuijnder <lennart.tuijnder@openanalytics.eu>
Authors@R: c(
	person(given = "Anne-Katrin", family = "Hess", email = "anne-katrin.hess@openanalytics.eu", role = c("aut")),
	person(given ="Lennart", family = "Tuijnder", email = "lennart.tuijnder@openanalytics.eu", role = c("aut", "cre")))
Description: Unit testing is a solid component of automated CI/CD pipelines. 
	'tinytest' - a lightweight, zero-dependency  alternative to 'testthat' was developed.
	To be able to integrate 'tinytests' results into common CI/CD systems the 'tinytests'-object
	is converted to JUnit XML format. 'tinytest2JUnit' enables this conversion 
	while staying lightweight, having only 'tinytest' as its dependency.
Imports: tools, parallel, utils, tinytest
URL: https://github.com/openanalytics/tinytest2JUnit
BugReports: https://github.com/openanalytics/tinytest2JUnit/issues
License: GPL-3
Copyright: Open Analytics NV, 2024
RoxygenNote: 7.2.3
Encoding: UTF-8
NeedsCompilation: no
Packaged: 2026-04-13 06:27:50 UTC; ltuijnder
Author: Anne-Katrin Hess [aut],
  Lennart Tuijnder [aut, cre]
Repository: CRAN
Date/Publication: 2026-04-13 06:50:02 UTC
Built: R 4.6.0; ; 2026-04-27 04:38:10 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method("[",tinytests2JUnit)
S3method(format,XMLtag)
S3method(print,XMLtag)
export(runTestDir)
export(testPackage)
export(writeJUnit)
importFrom(parallel,clusterCall)
importFrom(parallel,clusterEvalQ)
importFrom(parallel,clusterExport)
importFrom(parallel,makeCluster)
importFrom(tinytest,run_test_file)
importFrom(tinytest,tinytest)
importFrom(tools,file_path_sans_ext)
importFrom(utils,getFromNamespace)
importFrom(utils,getSrcDirectory)
importFrom(utils,getSrcFilename)
importFrom(utils,getSrcLocation)
importFrom(utils,tail)

#' Fibonacci 
#' 
#' Compute first n fibonacci numbers.
#'
#' @param n an integer
#' @return A numeric vector of length n
#' @export 
fibonacci <- function(n) {
  if (n == 20) stop("I am a bug!")
  fibs <- numeric(n)
  if (n >= 1) fibs[1] <- 1
  if (n >= 2) fibs[2] <- 1
  if (n > 2) for (i in 3:n) fibs[i] <- fibs[i - 1] + fibs[i - 2]
  return(fibs)
}

expect_equal(fibonacci(1), c(1))
expect_equal(fibonacci(2), c(1, 1))
expect_equal(fibonacci(3), c(1, 1, 2), info = "fibbonachi n=3 works")
expect_equal(fibonacci(4), c(1, 1, 2, 4), info = "A failing test!")

#' @importFrom parallel makeCluster clusterExport clusterCall clusterEvalQ
#' @importFrom utils getSrcFilename getSrcLocation getSrcDirectory tail getFromNamespace
#' @importFrom tools file_path_sans_ext
NULL

#' Help function to generate the formatted string for a single stack frame.
#'
#' Help function to generate the formatted string for a single stack frame.
#' 
#' @param framecall `character(n)`: deparsed call for the given stack. Each element corresponds to 
#'    a line.
#' @param frameN `integer(1)`: frame nummer.
#' @param hasSrcInfo `logical(1)`: Does the call have any source info? 
#' @param dirName `character(1)`: the directory name of the source file.
#' @param fileName `character(1)`: filename of the source. Value ignored if hasSrcInfo=TRUE.
#' @param lineNr `character(1)`: linenr in the source where the call occured..
#'    Value ignored if hasSrcInfo=TRUE.
#' 
#' @return `characer(1)` the formatted character string containing info of a single frame in the 
#'   stackstrace
#'
#' @details
#' For a given frame in the stack the string is formatted as follows (substitute the arguments 
#'  between the curly braces)
#' \code{
#' {frameN}| {call[1]}
#' {frameN}| {call[2]}
#' {frameN}| {call[3]}
#' {frameN}| {call[3]}
#' ---> at File={dirName/fileName} Line={line}:
#' }
#' 
#' For example for only a single line error:
#' \code{
#' 1: stop("This is a crash")
#' ---> at File=R/my_r_code_file.R Line=234
#' }
#'
#' Currently all call lines are printed for a given stack. 
#' The last line with source file info only printed if hasSrcInfo=TRUE. Else it is ommited. 
formattedFrame <- function(framecall, frameN, hasSrcInfo, dirName, fileName, lineNr) {
  prefixCall <- paste0(frameN, "| ")
  result <- paste(prefixCall, framecall, collapse = "\n")
  if (hasSrcInfo) {
    result <- paste0(
      result, "\n", 
      "---> at File=", file.path(dirName, fileName), " Line=", lineNr
    )
  }
  return(result)
}

#' Get formatted stack trace for an uncaught error from a tinytest test file.
#' 
#' getFormattedStacktrace is a helper function that formats stacktrace for uncaught errors from a 
#' tinytest run file. 
#' 
#' @details 
#' This function assumes that it directly called from a withCallingHandler 
#' error handling function! This fact is then used to remove the calling handler info from the 
#' stack such that stack directly starts from where the error was thrown.
#' 
#' The function also removes the calls from the stack that involve 
#' executing the `test_file`. The internals of runTestDir and tinytest are not of intererst. 
#' And the highest level of the stack to consider is the top level of the test_file.
#' 
#' Note, this does mean that errors that occur on the top-level of the test file will not have a 
#' a stacktrace!
#' For example: "Error: object 'x' not found" where x is attempted to be resolved 
#' at the root levels
#' 
#' @return `character(1)` a single length character string suitable to be printed to the end-user.
#'   In case of no stakctrace (eg the error occured at root level of the script) NA_character_ 
#'   is returned!
getFormattedStacktrace <- function() {
  trace <- .traceback(4)
  hasSrcInfo <- vapply(trace, function(frame) !is.null(attr(frame, "srcref")), logical(1))
  fileNames <- lapply(trace, utils::getSrcFilename)
  lineNr <- lapply(trace, utils::getSrcLocation)
  dirNames <- lapply(trace, utils::getSrcDirectory)
  result <- Map(
    f = formattedFrame,
    framecall = trace,
    frameN = seq_along(trace),
    hasSrcInfo = hasSrcInfo,
    dirName = dirNames,
    fileName = fileNames,
    lineNr = lineNr
  )

  # Remove parts of the stack that involve the internals of tinytest and tinytest2JUnit. 
  isFrameProvokingTest <- vapply(
    X = trace,
    FUN = function(frame) {
      # Note this stack depeneds on how we call the tinytest::run_test_file( !
      startsWith(as.character(frame[1]), "tinytest::run_test_file(file = file,")
    },
    FUN.VALUE = logical(1)
  )

  # If internals of tinytest change: simply default back to showing whole stacktrace
  if (any(isFrameProvokingTest)) {
    # There should be only 1 but you never known...
    indx <- utils::tail(which(isFrameProvokingTest), n = 1)

    # Remove tinytest internals. If these ever change: default to the whole stack
    if (indx > 2) {

      oneStackDeeper <- as.character(trace[[indx - 1]])
      twoStackDeeper <- as.character(trace[[indx - 2]])
      internalsTinytest <- "eval(expr, envir = e)"
      internalIsAsExpected <- (oneStackDeeper == internalsTinytest &&
        twoStackDeeper == internalsTinytest)
      if (internalIsAsExpected) {
        if (indx == 3) return(NA_character_) # No stacktrace
        result <- result[1:(indx - 3)] 
      }
    }
  }
  
  # Seperate the frame in the printing with a blank line. To improve readability
  paste0(
    vapply(X = result, FUN = `[[`, 1, FUN.VALUE = character(1)),
    collapse = "\n\n"
  )
}

#' tinytestJUnit test results
#'
#' An object of class `tinytests2JUnit`. Note the plurar. A subclass of [tinytest::tinytests()] 
#' containing extra info recordings that are used in the export to JUnit.
#'
#' @param x `tinytests2JUnit` object to subset. 
#' @param i object to subset the `x` with. 
#'
#' @details 
#' Following details are recorded when running the tests files and stored as additional attributes 
#' to the object:
#' 
#' * **fileDurations**: `named-numeric(n)`. Names = filename of tests files, value = duration 
#'   in seconds on how long the test file took to run.
#' * **fileTimestamps**: `named-character(n)`. Names = filename of tests files, value = timestamp 
#'   when the test was invoked.
#' * **fileHostnames**: `named-character(n)`. Names = filename of tests files, value = The hostname 
#'   of the system that ran the tests. (Usefull in combination with clusters). 
#' * **disabled**: `character`. A character vector of filenames where no tests were ran. 
#'    They are flagged as disabled tests.
#' 
#' @aliases tinytests2JUnit
#' 
#' @export 
#' @rdname tinytests2JUnit 
`[.tinytests2JUnit` <- function(x, i) {
  r <- unclass(x)[i]
  files <- unique(vapply(r, attr, "file", FUN.VALUE = character(1)))
  structure(
    r,
    class = c("tinytests2JUnit", "tinytests"),
    duration = NULL,
    fileDurations = attr(x, 'fileDurations')[files],
    fileTimestamps = attr(x, 'fileTimestamps')[files],
    fileHostnames = attr(x, 'fileHostnames')[files],
    disabled = attr(x, 'disabled')
  )
}

#' Internal wrapper arround tinytest::run_test_file
#' 
#' Internal wrapper arround [tinytest::run_test_file()] that records the test duration and 
#' catches uncaught errors and logs the stacktrace of where the error occured. 
#'
#' @details 
#' The response is a subclass of the `tinytests` object called: `tinytests2Junit` 
#' object which captures additional info for the reporting to JUnit:
#' 
#' * Duration to run the file.
#' * Timestamp when the test was run.
#' * hostname of the computer where it was ran on.
#' 
#' The caught error is turned into a subclass uncaught-error of tinytest. This is implementation 
#' detail and only to be understood by constructJUnitTag.
#'
#' If an error occured it is captured and `uncaught-error` object (subclass of `tinytest`) is 
#' returned in the `tinytests` object.
#' This tinytest object represents a "failed" tests that will get reported as an Error in the 
#' JUnit. Various aspects of the error are also captured like the the stacktrace.
#' 
#' @param file `character(1)` test file to run. 
#' @param ... arguments passed on to [tinytest::run_test_file()] 
#'
#' @return a `tinytests2JUnit` object (being a subclass of `tinytest` object).
#' @importFrom tinytest tinytest run_test_file
runTestFile <- function(file, ...) { 

  formattedStacktrace <- NA_character_
  timeStart <- Sys.time()
  testOutput <- tryCatch(
    # Catch safe the stacktrace where the uncaught error is signaled using 'withCallingHandlers'
    expr = withCallingHandlers(
      expr = tinytest::run_test_file(file = file, ...),
      error = function(e) formattedStacktrace <<- getFormattedStacktrace()
    ),
    error = function(e) {
      if (!is.null(e$call)) {
        errorPrint <- paste0("Error in ", deparse(e$call)[1], ": ", e$message)
      } else {
        errorPrint <- paste0("Error: ", e$message)
      }
      info <- paste(
        paste0("Uncaught error in test file: ", basename(file)),
        errorPrint,
        sep = "\n"
      )
      errorTest <- tinytest::tinytest(
        result = FALSE,
        call = e$call,
        diff = "No test captured since an error occured.",
        info = info,
        file = basename(file), 
        exception = e,
        formattedStacktrace = formattedStacktrace
      )
      class(errorTest) <- c("uncaught-error", "tinytest") # subclass
      return(structure(list(errorTest), class = "tinytests"))
    }
  )
  timeEnd <- Sys.time()

  # Take by preference the file defintion internally used by tinytest
  file <- if (length(testOutput) != 0) attr(testOutput[[1]], "file") else basename(file)

  fileDurations <- as.numeric(timeEnd) - as.numeric(timeStart)
  names(fileDurations) <- file 
  attr(testOutput, "fileDurations") <- fileDurations

  fileTimestamps <- strftime(timeStart, "%Y-%m-%dT%H:%M:%S%z")
  names(fileTimestamps) <- file
  attr(testOutput, "fileTimestamps") <- fileTimestamps

  fileHostnames <- Sys.info()['nodename']
  names(fileHostnames) <- file
  attr(testOutput, "fileHostnames") <- fileHostnames 

  attr(testOutput, "disabled") <- if (length(testOutput) == 0) file else character(0L)
  
  class(testOutput) <- c('tinytests2JUnit', class(testOutput))
  return(testOutput)
}

#' Run all the test files in a directory
#'
#' [runTestDir()] is a drop in replacement for [tinytest::run_test_dir()] with the key 
#' difference that errors thrown from within a test file are caught and get reported with a 
#' a stacktrace in the JUnit report. In addition, some extra metrics are recored for the JUnit 
#' report, such as: timestamp, test duration, hostname and if tests are disabled (see details for 
#' more info).
#' 
#' @param dir `character(1)` path to directory
#' @param at_home `logical(1)` should local tests be run? By default FALSE. Unlike 
#'   [tinytest::run_test_dir()] which is meant to be called in a local interactive context. This 
#'   function is meant to be called in a non-interactive CI environment, where we want to mimic the
#'   behaviour of how tests would get run by R CMD Check. 
#'   See also `at_home` documentation in tinytest package.
#' @param pattern `character(1)` A regular expression that is used to find
#'   scripts in \code{dir} containing tests (by default \code{.R} or \code{.r}
#'   files starting with \code{test}).
#' @param cluster A `cluster` object to run the test files on. Note, it is 
#'   expected that the clusters has already been prepared. Most notable, the package to test 
#'   should already been loaded. `runTestDir` will load the package 
#'   "tinytest" for you into the clusters. See [tinytest::run_test_dir()] for more details.
#' @param lc_collate See [tinytest::run_test_dir()].
#' @param ... Arguments passed on to [tinytest::run_test_file()]
#' 
#' @return A `tinytests2Junit` object to be provided to the [writeJUnit()] function.
#' 
#' @details
#' [runTestDir()] is meant as a CI-friendly alternative to the native [tinytest::run_test_dir()]. 
#' It catches errors that are raised in the tests files and adds them as a "failed" `tinytest` 
#' in the output. 
#'
#' [tinytest::run_test_dir()] would have let the error bubble up, stop the testing process and 
#' not report any failures from other tests. One is then also forced to look into the 
#' logs of the CI to see what the error was. The output of [runTestDir()] in combination with 
#' [writeJUnit()] will present you the error in the JUnit togheter with a stack trace. Next to the 
#' test results of the other files that ran without a problem. 
#' 
#' If you prefer the behaviour of [tinytest::run_test_dir()] you can still use it in combination 
#' [writeJUnit()].
#'
#' Caught errors are returned in the output as as sub-class of `tinytest` object. This is however 
#' considred implemenation detail and can be subject to change.
#' 
#' Note, function arguments explicilty listed in [tinytest::run_test_dir()] but not here can still
#' still be provided via `...`
#' 
#' @section tinytests2JUnit:
#' The returned object is a `tinytests2JUnit` object (note the plural). This object 
#' contains additional info compared to a `tinytests` object that is used in the JUnit report.
#' 
#' The following additional info will get reported:
#' 
#' * The timestamp per test file on when it got invoked. 
#' * The test duration per test file.
#' * The system hostname per test file on where it got invoked. This is mainly of interests for 
#'   different `clusters`. 
#' * If a test file is disabled. A test file is considered disabled if no tests 
#'   occur with in the file. 
#'   It is then assumed that at the top of file some conditional statement 
#'   made the test file exist early.
#'
#' @seealso 
#' * [tinytest::run_test_dir()] for how the function is inteded to behave.
#' * [writeJUnit()] where it is expected that the output of this function to be provided to.
#' * [testPackage()] for an higher-level function to simply test a package.
#' 
#' @export 
#' @examples 
#' # Run tests with `tinytest`
#' dirWithTests <- system.file("example_tests/multiple_files",package = "tinytest2JUnit")
#' testresults <- runTestDir(dirWithTests)
#' 
#' writeJUnit(testresults) # Writes content to stdout
runTestDir <- function(
  dir = "inst/tinytest", 
  at_home = FALSE,
  pattern = "^test.*\\.[rR]$",
  cluster = NULL,
  lc_collate = getOption("tt.collate", NA),
  ...
) {

  t0 <- Sys.time()

  testfiles <- dir(dir, pattern = pattern, full.names = TRUE)
  testfiles <- localeSort(testfiles, lc_collate)

  # If internals change do not crash
  try(setCallWd <- utils::getFromNamespace("set_call_wd", "tinytest")) 
  if (!inherits(cluster, "cluster")) {
    # set pwd here, to save time in run_test_file.
    oldwd <- getwd()
    try(setCallWd(oldwd)) 
    on.exit({
      setwd(oldwd)
      try(setCallWd(oldwd))
    })
    setwd(dir)  
    testOutput <- lapply(
      X = basename(testfiles),
      FUN = runTestFile,
      at_home = at_home,
      ...
    )
  } else {
    parallel::clusterEvalQ(cl = cluster, library(tinytest))           
    testOutput <- parallel::parLapply(
      cl = cluster,
      X = testfiles,
      fun = runTestFile,
      at_home = at_home,
      ...
    )
  }

  fileDurations <- unlist(lapply(testOutput, attr, 'fileDurations'), recursive = FALSE)
  fileTimestamps <- unlist(lapply(testOutput, attr, 'fileTimestamps'), recursive = FALSE)
  fileHostnames <- unlist(lapply(testOutput, attr, 'fileHostnames'), recursive = FALSE)
  disabled <- unlist(lapply(testOutput, attr, 'disabled'), recursive = FALSE)
  if (length(disabled) == 0) disabled <- character(0L) # Since: unlist(list()) -> NULL

  td <- abs(as.numeric(Sys.time()) - as.numeric(t0))
  structure(
    unlist(testOutput, recursive = FALSE),
    class = c("tinytests2JUnit", "tinytests"),
    duration = td,
    fileDurations = fileDurations,
    fileTimestamps = fileTimestamps,
    fileHostnames = fileHostnames,
    disabled = disabled 
  )
}


# Copied over from tinytest
localeSort <- function(x, lcCollate = NA, ...) {
  if (is.na(lcCollate)) return(sort(x, ...))

  # catch current locale
  oldCollate <- Sys.getlocale("LC_COLLATE")

  # set to user-defined locale if possible, otherwise sort using current locale 
  colset <- tryCatch(
    expr = {
      Sys.setlocale("LC_COLLATE", lcCollate)
      TRUE
    }, 
    warning = function(e) { 
      msg <- sprintf("Could not sort test files in 'C' locale, using %s\n", oldCollate)
      message(paste(msg, e$message, "\n")) 
      FALSE
    }, 
    error = warning
  )
  out <- sort(x)

  # reset to old locale
  if (colset) Sys.setlocale("LC_COLLATE", oldCollate)
  out
}

foldersToTest <- system.file("example_tests", package = "tinytest2JUnit") |>
  list.dirs(full.names = TRUE, recursive = FALSE)

folderName <- system.file("example_tests", package = "tinytest2JUnit") |>
  list.dirs(full.names = FALSE, recursive = FALSE)

# Test that writeJUnit can deal with all the test folders defined in example_tests
for (i in seq_along(foldersToTest)) {
  
  tmpFile <- tempfile(fileext = ".xml")
  if (file.exists(tmpFile)) stop("File should not have existed!")
  
  testresults <- runTestDir(foldersToTest[i], verbose = FALSE)

  expect_true(
    is.logical(writeJUnit(tinytests = testresults, file = tmpFile)),
    info = paste0("writeJUnit() succeeded for directory: ", folderName[i])
  )
  expect_true(file.exists(tmpFile))
  expect_equal(
    readLines(tmpFile, n = 1),
    "<?xml version=\"1.0\" encoding=\"UTF-8\"?>"
  )
}

# Test that when overwrite = TRUE. The file is actually being overwritten.

tmpFile <- tempfile(fileext = ".xml")
if (file.exists(tmpFile)) stop("File should not have existed!")

randomTestFolder <- sample(foldersToTest, size  = 1)
testresults <- runTestDir(randomTestFolder, verbose = FALSE)

writeJUnit(tinytests = testresults, file = tmpFile)
nLinesWrittenFirstTime <- length(readLines(tmpFile))

writeJUnit(tinytests = testresults, file = tmpFile, overwrite = TRUE)
nLinesWrittenSecondTime <- length(readLines(tmpFile))

expect_equal(
  nLinesWrittenFirstTime,
  nLinesWrittenSecondTime,
  info = "Overwrite = TRUE does overwrite the file"
)

expect_error(
  writeJUnit(tinytests = testresults, file = tmpFile, overwrite = FALSE),
  info = "Overwrite = FALSE does indeed not overwrite a file if it already exists."
)

expect_equal(
  length(utils::capture.output(writeJUnit(testresults))),
  nLinesWrittenFirstTime, 
  info = "writeJUnit by defaults write to stdout"
)

testResults <- runTestDir(
  system.file("example_tests/crashing_tests", package = "tinytest2JUnit"),
  verbose = 0
)

expect_inherits(testResults, "tinytests2JUnit", info = "tinytests2JUnit is returned")
expect_inherits(testResults, "tinytests", info = "tinytests is returned")
expect_true(
  all(vapply(testResults, inherits, "tinytest", FUN.VALUE = logical(1))),
  info = "all items are tinytest"
)
expect_false(all(unlist(testResults)), info = "Crashing tests is reported as failin tinytest")

tags <- tinytest2JUnit:::constructTestsuitesTag(testResults)
expect_equal(
  attr(tags, "class"),
  "XMLtag",
  info = "constructTestsuitesTag works with the results from runTestDir."
)
expect_equal(tags$attributes$tests, 1L, info = "Crashing test is reported")
expect_equal(tags$attributes$failures, 0L, info = "Crashing test is not reported as failure")
expect_equal(tags$attributes$errors, 1L, info = "Crashing test is reported as error")
expect_equal(tags$content[[1]]$attributes$errors, 1L, info = "Crashing test is reported as error")

testDir <- system.file("example_tests/multiple_files", package = "tinytest2JUnit")
files <- sort(dir(testDir))
testResults <- runTestDir(testDir, verbose = 0)
expect_inherits(testResults, "tinytests2JUnit", info = "tinytests2JUnit is returned")

vctFiles <- vapply(testResults, attr, "file", FUN.VALUE = character(1L))
splitPerFile <- lapply(files, function(f) testResults[vctFiles == f])
expect_true(
  all(vapply(splitPerFile, inherits, 'tinytests2JUnit', FUN.VALUE = logical(1))),
  info = 'Subsetting tinytests2JUnit yields tinytests2JUnit'
)

# Test file duration property:
expect_true(is.numeric(attr(testResults, "fileDurations")), info = "File durations is reported")
expect_equal(
  files,
  names(attr(testResults, "fileDurations")),
  info = "Duration of all files are reported."
)
expect_equal(
  names(attr(splitPerFile[[1]], "fileDurations")),
  target = files[1],
  info = "After subset duration of only relevant files is still reported."
)
expect_equal(
  attr(splitPerFile[[2]][[1]], 'file'),
  names(attr(splitPerFile[[2]], "fileDurations")),
  info = "Name fileDurations is equal to the file attribute of the tinytest object"
)

# Test file timestamp
expect_true(
  is.character(attr(testResults, "fileTimestamps")),
  info = "File timestamps is reported"
)
expect_equal(
  names(attr(splitPerFile[[2]], "fileTimestamps")),
  target = files[2],
  info = "fileTimestamps is present after subsetting and only for relevant files."
)

# Test file hostnames
expect_equal(
  names(attr(splitPerFile[[2]], "fileHostnames")),
  target = files[2],
  info = "fileHostnames is present after subsetting and only for relevant files."
)

# Test disabled.
expect_equal(attr(testResults, 'disabled'), attr(splitPerFile[[1]], 'disabled'))


tagsMultipleFiles <- tinytest2JUnit:::constructTestsuitesTag(testResults)
expect_true(
  tagsMultipleFiles$attributes$tests == 3L,
  info = "runTestDir also works with non-crashing tests."
)
expect_true(
  tagsMultipleFiles$attributes$failures == 1L,
  info = "runTestDir also still correctly reports normal failing tests."
)
expect_true(
  tagsMultipleFiles$attributes$disabled == 0L,
  info = "tinytests2JUnit results in disabled reporting"
)
expect_true(
  !is.null(tagsMultipleFiles$content[[1]]$attributes$time),
  info = "time is reported for tinytests2JUnit objects"
)
expect_true(
  !is.null(tagsMultipleFiles$content[[1]]$attributes$hostname),
  info = "hostname is reported for tinytests2JUnit objects"
)
expect_true(
  !is.null(tagsMultipleFiles$content[[1]]$attributes$timestamp),
  info = "timestamp is reported for tinytests2JUnit objects"
)

testResults <- runTestDir(
  system.file("example_tests/get_call_wd", package = "tinytest2JUnit"),
  verbose = 2
)
testResults <- tinytest::run_test_dir(
  system.file("example_tests/get_call_wd", package = "tinytest2JUnit"),
  verbose = 2
)

if (at_home()) { # Problems on windows computer. 
  expect_true(
    all(unlist(testResults)),
    info = "runTestDir correctly sets working directory and calling directory"
  )
}

# Test disabled tests are being reported:
testDir <- system.file("example_tests/skips", package = "tinytest2JUnit")
testResults <- runTestDir(testDir, verbose = 0)
expect_equal(attr(testResults, "disabled"), "test_everything_skipped.R")
tags <- tinytest2JUnit:::constructTestsuitesTag(testResults)
expect_equal(tags$attributes$disabled, 1L)


cluster <- try(parallel::makeCluster(2))
if (inherits(cluster, "try-error")) {
  exit_file("Cannot create cluster. Skipping cluster tests.")
}

tryCatch(
  resultCluster <- runTestDir(
    system.file("example_tests/multiple_files", package = "tinytest2JUnit"),
    cluster = cluster,
    verbose = 0
  ),
  finally = parallel::stopCluster(cluster)
)

expect_true(exists("resultCluster"), info = "Cluster running is supported")
tagsMultipleFiles <- tinytest2JUnit:::constructTestsuitesTag(resultCluster)
expect_true(
  length(resultCluster) == 3L,
  info = "runTestDir correctly runs with clusters."
)

#' Test an R package and report the results in JUnit 
#'
#' Run all tests of a package and report the results as JUnit xml. This function 
#' can be seen as a drop in replacement for [tinytest::test_package()] but with a 
#' key difference that uncaught errors will be catched and reported JUnit!
#' This function is intended to be used in a test stage of a CI build.
#' 
#' @param pkgname `character(1)`. Name of the package to tests. 
#' @param file `character(1) | connection`: Full file path or connection object to write the 
#'   JUnit xml content to. By default `stdout()` connection is used.
#'   **Warning** if the file already exist it will be overwritten!
#' @param errorOnFailure `logical(1)` Should an error be raised (after writing the JUnit) when a 
#'   at least one test failed? By default TRUE. This is done as a 
#'   convenience to have the CI fail at the test stage on failure.
#' @param testdir `character(1)` testing directory of the package. See ?[tinytest::test_package()] 
#'   for more details.
#' @param lib.loc `character(1) | NULL` Library location where the package is installed. 
#'   By default: `NULL` meaning the package is searched on the standard .libPaths().
#' @param at_home `logical(1)` should local test be run? By default FALSE as we want to as closely 
#'    mimic the environement of how tests would get ran in CRAN. 
#'    See for more details [tinytest::test_package()].
#' @param ncpu `postive integer(1) | clutser` Either an integer specifying the amount of cpu's 
#'   to parralize the testing over or a `cluster` object to run the tests in.
#' @param ... Extra arguments passed on to [runTestDir()]
#' 
#' @return If `errorOnFailure` = FALSE, a `tinytests2JUnit` object (a subclass of `tinytests` 
#'   object that captures more info for export to JUnit). Else, an error is raised if at least 
#'   on failure occurs. Meant as convenience to automatically stop the CI build.
#'
#' @details 
#' [testPackage()] is meant as a CI-friendly alternative to the native [tinytest::test_package()]. 
#' Next to directly reporting the tests results in a JUnit xml format, it also catches errors that 
#' are raised in the tests files and reports them as "error" in the JUnit. 
#'
#' [tinytest::test_package()] would have let the error bubble up, stop the testing 
#' process and not report any failures from other test files. One is then also forced 
#' to look into the logs of the CI to see what the error was. [testPackage()] presents you that
#' error in the JUnit with a stacktrace. Next to all the test results of the other files that 
#' ran without a problem. 
#' 
#' If you prefer the behaviour from [tinytest::test_package()], you can still use it in 
#' combination with [writeJUnit()] if all tests results pass.
#'  
#' Just like [tinytest::test_package()] an error is raised if at least one failure occured during 
#' testing. Obviously catched errors are also seen as failures. This error is raised
#' **after** the test results have been written away to the file, such that your CI can still pick 
#' it up and report the failure.
#' The error raising is done as a convenience to stop the CI from continue if test-failure occured.
#' You can opt-out of this behaviour by setting the `errorOnFailure` parameter to FALSE. Then a
#' case `tinytests2JUnit` object is returned (a sub-class of `tinytests` object containing addition 
#' info for the JUnit). 
#' Caught errors are also captured in this object as `tinytest`-objects. They 
#' actually have a special sub-class but this is considered an internal implemenation detail.
#' 
#' [testPackage()] is NOT meant to be called from within your `tests/tinytests.R` file! Tests 
#' invoked by R CMD Check or on CRAN should still make use of [tinytest::test_package()].
#' This function is only meant to be called from within a testing step in your CI to 
#' report the test results in an JUnit xml format.
#' 
#' @inheritSection writeJUnit Side-effects
#' @inheritSection writeJUnit tinytests to JUnit
#' @export
#' @seealso [runTestDir()] and [tinytest::test_package()].
#' @examples
#' tmpFile <- tempfile(fileext = ".xml")
#' testPackage("tinytest", file = tmpFile, verbose = 0)
testPackage <- function(
  pkgname,
  file = stdout(),
  errorOnFailure = TRUE, 
  testdir = "tinytest",
  lib.loc = NULL,
  at_home = FALSE,
  ncpu = NULL,
  ...
) {
  oldlibpaths <- .libPaths()

  if (!is.null(lib.loc)) { 
    e <- new.env()
    e$libs <- c(lib.loc, oldlibpaths)
      
    if (!dir.exists(lib.loc)) 
      warning("lib.loc '", lib.loc, "' not found.")
    .libPaths(c(lib.loc, oldlibpaths))
  }

  on.exit({
    if (is.numeric(ncpu)) parallel::stopCluster(cluster)
    .libPaths(oldlibpaths)
  })

  if (!dir.exists(testdir)) { # if not customized test dir
    # find the installed test dir
    newTestDir <- system.file(testdir, package = pkgname, lib.loc = lib.loc)
    if (newTestDir == "") {
      stop("testdir '", testdir, "' not found for package '", pkgname, "'")
    } else {
      testdir <- newTestDir
    }
  }

  # set up cluster if required
  cluster <- if (is.null(ncpu)) NULL
  if (is.null(ncpu)) {
    cluster <- NULL 
  } else if (is.numeric(ncpu)) {
    cluster <- parallel::makeCluster(ncpu, outfile = "")
  } else if (inherits(ncpu, "cluster")) {
    cluster <- ncpu
  } else {
    stop("ncpu must be NULL, 'numeric', or 'cluster'")
  }

  # By now we have a cluster, or NULL. Load the pkg under scrutiny.
  if (is.null(cluster)) {
    library(pkgname, character.only = TRUE, lib.loc = lib.loc)
  } else {
    if (!is.null(lib.loc)) {
      # to prevent a R CMD check warning we must create a dummy libs here
      # as well
      libs <- NULL
      parallel::clusterExport(cluster, "libs", envir = e)
      parallel::clusterEvalQ(cluster, .libPaths(libs))
    }
    parallel::clusterCall(cluster, library, pkgname, character.only = TRUE, lib.loc = lib.loc)
  }

  out <- runTestDir(testdir, at_home = at_home, cluster = cluster, ...) 
  writeJUnit(tinytests = out, file = file, overwrite = TRUE)
  vTestFailed <- sapply(out, isFALSE)
  if (any(vTestFailed) && errorOnFailure) {
    writeLines(vapply(out[vTestFailed], format, type = "long", FUN.VALUE = character(1)))
    stop(sum(vTestFailed), " out of ", length(out), " tests failed", call.= FALSE)
  } 
  out
}

tmpFile <- tempfile(fileext = ".xml")
expect_inherits(
  testPackage("tinytest", file = tmpFile, verbose = 0),
  "tinytests2JUnit",
  info = "Works with package in standard lib path."
)
startFile <- readLines(tmpFile, n = 1)
expect_equal(startFile, '<?xml version="1.0" encoding="UTF-8"?>')


tmpLib <- tempdir()
fibPackage <- system.file(
  "example_package/fib_1.0.0.tar.gz",
  package = "tinytest2JUnit",
  mustWork = TRUE
)
install.packages(fibPackage, type = "source", repos = NULL, lib = tmpLib)

testDir <- system.file("example_package/fib/inst/tinytest", package = "tinytest2JUnit")
expect_inherits(
  result <- testPackage(
    "fib",
    file = tmpFile, 
    testdir = testDir,
    verbose = 0,
    lib.loc = tmpLib,
    errorOnFailure = FALSE
  ),
  "tinytests2JUnit",
  info = "Works with a directly provided path and custom lib and errorOnFailure = FALSE."
)
expect_equal(sum(sapply(result, isFALSE)), 2L, info = "2 failed tinytests")

expect_inherits(
  testResults <- testPackage(
    "fib",
    file = tmpFile, 
    ncpu = 2,
    errorOnFailure = FALSE,
    lib = tmpLib,
    verbose = 0
  ),
  "tinytests2JUnit",
  info = "Parralization works"
)

tags <- tinytest2JUnit:::constructTestsuitesTag(testResults)
expect_equal(tags$attributes$errors, 1L, info = "The hard crash is correctly registed as an Error.")


cluster <- parallel::makeCluster(2)
tryCatch(
  expr = {
    testResults <- testPackage(
      "fib",
      file = tmpFile, 
      ncpu = cluster,
      errorOnFailure = FALSE,
      lib.loc = tmpLib
    )
    expect_inherits(
      testResults,
      "tinytests2JUnit",
      info = "Custom clusters works."
    )
  },
  finally =  parallel::stopCluster(cluster)
)


expect_error(
  testPackage("fib", file = tmpFile, lib.loc = tmpLib),
  info = "By default an error is raised when a test failure occurs"
)

#' Test if single length character non NA.
#' @param x object to test.
#' @return `logical(1)`
#' @author ltuijnder
isSingleLengthCharNonNA <- function(x) {
  is.character(x) && length(x) == 1 && !is.na(x)
}


#' Convert any will character vector to a single length character vector 
#'
#' @param x a `character`
#' @return x a single-length character vector Non-NA
#' @examples
#' tinytest2JUnit:::charVecToSingleLength(c("Hello", "World")) # -> "HelloWorld"
#' tinytest2JUnit:::charVecToSingleLength(c("Hello", NA_character_)) # -> "HelloNA"
#' tinytest2JUnit:::charVecToSingleLength(character(0L)) # -> ""
charVecToSingleLength <- function(x) {
  if (length(x) == 0) return("")
  x[is.na(x)] <- "NA"
  if (length(x) == 1) return(x)
  return(paste0(x, collapse = ""))
}

testResults <- tinytest::run_test_dir(
  dir = system.file("example_tests/empty_tests", package = "tinytest2JUnit"),
  verbose = FALSE
)
tags <- tinytest2JUnit:::constructTestsuitesTag(testResults)
expect_equal(
  attr(tags, "class"),
  "XMLtag",
  info = "constructTestsuitesTag returns a xml tag object"
)
expect_equal(tags$attributes$tests, 0L, info = "empty test file results in 0 tests")
expect_true(
  all(c("name", "tests", "failures", "errors", "time", "disabled") %in% names(tags$attributes)), 
  info = "Expected testsuites tag attributes are present"
)
expect_true(!is.na(tags$attributes$time), info = "Total run time is reported.")
expect_true(
  tinytest2JUnit:::isSingleLengthCharNonNA(tags$attributes$name),
  info = "Testsuites name is set"
)


testResults <- tinytest::run_test_dir(
  system.file("example_tests/heavy_calculations", package = "tinytest2JUnit"), verbose = FALSE
)
tags <- tinytest2JUnit:::constructTestsuitesTag(testResults)
expect_true(tags$attributes$time > 1, info = "Time duration is correctly reported")
expect_true(tags$attributes$tests == 1L, info = "Tests are reported")
expect_equal(tags$content[[1]]$name, "testsuite", info = "testsuite tag is present.")

testResults <- tinytest::run_test_dir(
  dir = system.file("example_tests/multi_line_except_statement", package = "tinytest2JUnit"), 
  verbose = FALSE
)
tags <- tinytest2JUnit:::constructTestsuitesTag(testResults)
expect_equal(
  attr(tags, "class"),
  "XMLtag",
  info = "multi line except statements does not cause problems."
)

testResults <- tinytest::run_test_dir(
  dir = system.file("example_tests/multiple_files", package = "tinytest2JUnit"), 
  verbose = FALSE
)
tags <- tinytest2JUnit:::constructTestsuitesTag(testResults)
expect_true(tags$attributes$tests == 3L)
expect_true(tags$attributes$failures == 1L)
expect_equal(
  vapply(
    tags$content,
    function(testsuite) testsuite$attributes$name,
    FUN.VALUE = character(1)
  ),
  target = c("test_file1", "test_file2"),
  info = "Name of testsuite-tag is equal to the test_file without suffix."
)
expect_equal(
  tags$content[[1]]$content[[1]]$attributes, 
  list(classname = "test_file1", name = "test_file1.R L9", status = "PASSED"),
  info = "testcase name and status are correctly set"
)
expect_equal(
  tags$content[[1]]$content[[2]]$attributes, 
  list(classname = "test_file1", name = "test_file1.R L10 : Some info", status = "PASSED"),
  info = "Info get's added to the the testcase name."
)
expect_equal(
  tags$content[[2]]$content[[1]]$attributes, 
  list(classname = "test_file2", name = "test_file2.R L9-L15", status = "FAILED"),
  info = "Failed testcase are handled!"
)


testResults <- tinytest::run_test_dir(
  dir = system.file("example_tests/skips", package = "tinytest2JUnit"),
  verbose = FALSE
)
tags <- tinytest2JUnit:::constructTestsuitesTag(testResults)
expect_true(tags$attributes$tests == 1L, info = "Skipped tests are not reported.")


testResults <- tinytest::run_test_dir(
  dir = system.file("example_tests/side_effects", package =  "tinytest2JUnit"),
  verbose = FALSE
)
tags <- tinytest2JUnit:::constructTestsuitesTag(testResults)
expect_true(tags$attributes$tests == 3L, info  = "Side effect is reported")
# TODO if side-effect is correctly reported.

isSingleLengthCharNonNA <- tinytest2JUnit:::isSingleLengthCharNonNA
charVecToSingleLength <- tinytest2JUnit:::charVecToSingleLength

expect_true(
  isSingleLengthCharNonNA("!"),
  info = "isSingleLengthIntNonNA works with valid characters"
)
expect_true(
  isSingleLengthCharNonNA(""),
  info = "isSingleLengthIntNonNA works ''"
)

expect_false(
  isSingleLengthCharNonNA(NA),
  info = "isSingleLengthIntNonNA correctly falsifies with NA"
)
expect_false(
  isSingleLengthCharNonNA(NA_character_),
  info = "isSingleLengthIntNonNA correctly falsifies with NA"
)
expect_false(
  isSingleLengthCharNonNA(NaN),
  info = "isSingleLengthIntNonNA correctly falsifies with NA"
)


expect_false(
  isSingleLengthCharNonNA(c("A", "B")),
  info = "isSingleLengthIntNonNA is false for length != 1"
)
expect_false(
  isSingleLengthCharNonNA(c(NA_character_, NA_character_)),
  info = "isSingleLengthIntNonNA is false for length != 1"
)
expect_false(
  isSingleLengthCharNonNA(character(0L)),
  info = "isSingleLengthIntNonNA is false for length != 1"
)


expect_equal(
  charVecToSingleLength("Hello\nWorld"),
  "Hello\nWorld",
  info = "Leaves normal length-1 alone"
)

expect_equal(
  charVecToSingleLength(c("Hello", "World")),
  "HelloWorld",
  info = "charVecToSingleLength concates them with empty space"
)

expect_equal(
  charVecToSingleLength(c("Hello", NA_character_, "World")),
  "HelloNAWorld",
  info = "NA is converted is converted to a literal NA"
)

expect_equal(
  charVecToSingleLength(character(0L)),
  "",
  info = "Zero-length is converted to empty string."
)

#' Write the results of a `tinytests`-object into JUnit xml report.
#' 
#' Write the `tinytests`-object to a JUnit XML reporting file. If a `tinytests2JUnit` is provided 
#' (returned by [runTestDir()]) more info will get reported.
#' 
#' @param tinytests `tinytests`-object to convert to JUnit xml. 
#' @param file `character(1) | connection`: Full file path or connection object to write the 
#'  JUnit xml content to. By default `stdout()` connection is used.
#' @param overwrite `logical(1)`: should the file be overwritten if it already exist? 
#'  By default TRUE.
#' 
#' @return `invisible(TRUE)` Might get another use in the future.
#' 
#' @seealso The JUnit XML report format:  \url{https://github.com/testmoapp/junitxml}
#' 
#' @section Errors:
#' In case of overwrite = FALSE and the file already exists an error is thrown.
#' 
#' @section Side-effects:
#' Side effects are registered as 'passed' tests in the JUnit output and have been given a status 
#' "SIDE-EFFECT". The call and diff is also returned in the standard-output of the testcase tag.
#' 
#' They are not considred failures and would thus not stop a pipeline.
#' 
#' @section tinytests to JUnit:
#' 
#' To comply the the JUnit specification the tests results are adapted as follows:
#'
#' * A single test run `tinytests` is mapped to a `<testsuites>` tag. 
#' * All `tinytest` results from a single file are mapped to a single `<testsuite>` tag. 
#'   * The name of the testsuite is equal to the test file name (without the file suffix)
#' * An individual `tinytest` object (eg. a single `except_*` exception test) is mapped to a 
#'   `<testcase>` tag. 
#'   * The name of the testcase is equal to the fileName + Line specification of where the expect
#'     statement is performed + the info.
#'
#' For reference: \url{https://github.com/testmoapp/junitxml}
#' 
#' @export
#' @examples 
#' # Run tests with `tinytest`
#' dirWithTests <- system.file("example_tests/multiple_files",package = "tinytest2JUnit")
#' testresults <- runTestDir(dirWithTests)
#' 
#' writeJUnit(testresults) # Writes content to stdout
#' 
#' tmpFile <- tempfile(fileext = ".xml")
#' writeJUnit(tinytests = testresults, file = tmpFile)
writeJUnit <- function(tinytests, file = stdout(), overwrite = TRUE) {
  
  if (!isSingleLengthCharNonNA(file) && !inherits(file, "connection")) {
    stop("File should be single length non-NA character or a connection object.")
  }
  if (is.character(file)) {
    fileExists <- file.exists(file)
    if (fileExists && !overwrite) {
      stop("Overwrite is set to FALSE and specified file already exists: ", file)
    }
  }
  
  junitXML <- constructTestsuitesTag(testResults = tinytests)
  cat('<?xml version="1.0" encoding="UTF-8"?>\n', file = file, append = FALSE)
  cat(format(junitXML), sep = "\n", file = file, append = TRUE)
  
  invisible(TRUE)
}

#' Construct the JUnit `</testsuites>` tag
#' 
#' Convert the `tinytests2Junit` or `tinytests`-object containing test across 
#' possibly multiple files into a JUnit `</testsuites>` tag. More details are reported to the 
#' JUnit if a `tinytests2JUnit` object compared to the native `tinytests` object.
#' 
#' @details
#' Reference for JUnit XML format: https://llg.cubic.org/docs/junit/
#'
#' See details [runTestDir()] which additional info is recorded.
#' 
#' @param testResults `tinytests2Junit | `tinytests` object to convert into a JUnit XML object 
#'   Usually the result of [runTestDir()] for a `tinytests2JUnit` object or native `tinytests`
#'   produced by [tinytest::test_package()] or [tinytest::run_test_dir()].
#'
#' @return `XMLtag`: with tag-name = `</testsuites>`. This is the root of the JUnit XML document.
constructTestsuitesTag <- function(testResults) {
 
  stopifnot(inherits(testResults, "tinytests"))
  
  vctTestFiles <- vapply(
    testResults,
    function(tinytest) attr(tinytest, "file"),
    FUN.VALUE = character(1)
  )
  isFailure <- vapply(
    testResults,
    function(tinytest) isFALSE(tinytest) && !inherits(tinytest, "uncaught-error"),
    FUN.VALUE = logical(1)
  )
  isError <- vapply(
    testResults,
    function(tinytest) isFALSE(tinytest) && inherits(tinytest, "uncaught-error"),
    FUN.VALUE = logical(1)
  )
  
  attributes <- list(
    name = "tinytest results", 
    tests = length(testResults), 
    failures = sum(isFailure, na.rm = TRUE),
    errors = sum(isError, na.rm = TRUE),
    disabled = length(attr(testResults, "disabled"))
  )
  
  duration <- attr(testResults, "duration")
  if (is.numeric(duration) && !is.na(duration)) attributes$time <- as.character(duration)
  
  uFiles <- unique(vctTestFiles)
  testsuiteTags <- Map(
    f = constructTestsuiteTag,
    testsFile = lapply(uFiles, function(f) testResults[vctTestFiles == f]),
    id = seq_along(uFiles) - 1 # JUnit uses 0 based index.
  )
  tag(
    "testsuites",
    attributes = attributes,
    content = testsuiteTags
  )
}

#' Construct JUnit `</testsuite>` tag 
#' 
#' Construct the `</testsuite>` tag of a `tinytest`, given all the `tinytest` results 
#' from a single test file. 
#'
#' @details
#' In case a `tinytest2JUnit` is provided following additional info can be reported:
#' 
#' * testsuite duration.
#' * timestamp when the testsuite was performed.
#' * hostname where the testsuite was ran.
#' 
#' @param testsFile `tinytests2JUnit | tinytests` -object with all test 
#'    results of a specified test file. At least a single test is expected.
#' @param id `integer(1)` testsuite id.
#' 
#' @return `XMLtag`: with tag-name = `</testsuite>` 
#'   that contains all the test results per test file.
constructTestsuiteTag <- function(testsFile, id) {
  
  stopifnot(inherits(testsFile, "tinytests"))
  
  isFailed <- vapply(
    testsFile,
    function(tinytest) isFALSE(tinytest) && !inherits(tinytest, "uncaught-error"),
    FUN.VALUE = logical(1)
  )
  isError <- vapply(
    testsFile,
    function(tinytest) isFALSE(tinytest) && inherits(tinytest, "uncaught-error"),
    FUN.VALUE = logical(1)
  )
  fileName <- attr(testsFile[[1]], "file")
  fileNameWithoutExt <- tools::file_path_sans_ext(fileName) 
  attributes <- list(
    name = escapeXml(fileNameWithoutExt),
    id = id,
    tests = length(testsFile),
    failures = sum(isFailed, na.rm = TRUE),
    errors = sum(isError, na.rm = TRUE)
  )

  if (inherits(testsFile, 'tinytests2JUnit')) {
    attributes$hostname <- escapeXml(attr(testsFile, "fileHostnames")[fileName])
    attributes$time <- as.character(attr(testsFile, "fileDurations")[fileName])
    attributes$timestamp <- escapeXml(attr(testsFile, "fileTimestamps")[fileName])
  }
  
  tag(
    "testsuite",
    attributes = attributes,
    content = lapply(testsFile, constructTestcaseTag)
  )
}


#' Construct JUnit `</testcase>` tag 
#' 
#' Construct JUnit `</testcase>` tag based on a single `tinytest` result.
#' 
#' @param tinytest a `tinytest`-object representing an individual test case. 
#' @return `XMLtag`: with tag-name = `tinytest` and contains the test result per test.
constructTestcaseTag <- function(tinytest) {
  
  stopifnot(inherits(tinytest, "tinytest"))

  if (isTRUE(tinytest)) {
    testcaseTag <- passedTestcaseTag(tinytest)   
  } else if (isFALSE(tinytest) && !inherits(tinytest, "uncaught-error")) {
    testcaseTag <- failureTestcaseTag(tinytest)
  } else if (isFALSE(tinytest) && inherits(tinytest, "uncaught-error")) {
    testcaseTag <- errorTestcaseTag(tinytest)
  } else { # tinytest = NA = side-effect
    testcaseTag <- sideeffectTestcaseTag(tinytest)
  }
  return(testcaseTag)
}

#' Helper function to construct the name of a testcase 
#'
#' Helper function to construct the name of a testcase. Note, the charater is already xml escaped.
#' 
#' @param tinytest a `tinytest` object. (does not matter what result)
#' @return `character(1)` the testcase name to use for this tinytest object.
nameTestcase <- function(tinytest) {
  
  file <- attr(tinytest, "file")
  fst <- attr(tinytest, "fst")
  lst <- attr(tinytest, "lst")
  info <- attr(tinytest, "info")

  # Name = {file}L{fst}-L{lst} : {info}
  name <- file
  if (!is.na(fst)) {
    name <- paste0(name, " L", fst)
  }
  if (!is.na(lst) && (fst != lst)) {
    name <- paste0(name, "-L", lst)
  }
  if (!is.na(info)) {
    infoSplit <- strsplit(info[1], "\n")[[1]]
    name <- paste0(name, " : ", infoSplit[1])
  }
  if (nchar(name) == 0) name <- "unnamed"
  name <- escapeXml(name)
  return(name)
}

#' Helper function specifying the 'classname' attribute of the testcase tag.
#' 
#' Helper function specifying the 'classname' attribute of the testcase tag. 
#' Currently equal to the fileName. The classname is already xml escaped.
#' 
#' @param tinytest a `tinytest`-object representing an individual test case. 
#' @return `character(1)` being the 'classname'
classnameTestcase <- function(tinytest) {
  file <- attr(tinytest, "file")
  fileName <- tools::file_path_sans_ext(file) 
  escapeXml(fileName)
}

#' Construct a testcase-tag for a passed test.
#' 
#' @param tinytest a `tinytest` object already validated to be a "PASSED" test.
#' @return a `testase` `XMLtag`
passedTestcaseTag <- function(tinytest) {
  testcaseTag <- tag(
    name = "testcase",
    attributes = list(
      classname = classnameTestcase(tinytest),
      name = nameTestcase(tinytest),
      status = "PASSED"
    )
  )
  return(testcaseTag)
}


#' Helper function generating the body of a failure description tag!
#'
#' Helper function generating the body of a failure description tag! Attempts to mimic the print 
#' behaviour of a tinytest object.
#'
#' @param tinytest A `tinytest` objec that is considered failed!.
#' @return `character(1)` being the failure tag description body. This string is already propery
#'   xml escaped.
constructFailureDescription <- function(tinytest) {
  call <- attr(tinytest, "call")
  if (is.language(call)) {
    callDesc <- paste0('call| ', deparse(call))
  } else {
    callDesc <- character(0L)
  }

  diff <- attr(tinytest, "diff")
  if (!is.na(diff)) {
    diffDesc <- paste0('diff| ', diff) 
  } else {
    diffDesc <- character(0L)
  }
  
  info <- attr(tinytest, "info")
  if (!is.na(info)) {
    infoSplit <- strsplit(info[1], "\n")[[1]]
    infoDesc <- paste0('info| ', infoSplit) 
  } else {
    infoDesc <- character(0L)
  }

  failureDescription <- escapeXmlText(paste0(c(callDesc, diffDesc, infoDesc), collapse = "\n"))
  failureDescription
}

#' Construct a testcase-tag for a failed test.
#' 
#' @param tinytest a `tinytest` object already validated to be a "FAILURE" test.
#' @return a `testase` `XMLtag`
failureTestcaseTag <- function(tinytest) {
  
  short <- attr(tinytest, "short")
  diff <- attr(tinytest, "diff")
  failureTag <- tag(
    name = "failure",
    attributes = list(
      message = if (!is.na(diff)) escapeXml(diff[1]) else "",
      type = if (!is.na(short)) short else "failure"
    ),
    content = list(
      constructFailureDescription(tinytest)
    )
  )
  testcaseTag <- tag(
    name = "testcase",
    attributes = list(
      classname = classnameTestcase(tinytest),
      name = nameTestcase(tinytest),
      status = "FAILED"
    ),
    content = list(failureTag)
  )
  return(testcaseTag)
}

#' Construct a testcase-tag for an error test.
#'
#' @param tinytest a `tinytest` object already validated to be a "ERROR" test.
#' @return a `testase` `XMLtag`
errorTestcaseTag <- function(tinytest) {
  
  e <- attr(tinytest, "exception")
  if (!is.null(e$call)) {
    errorMsg <- paste0("Error in ", deparse(e$call[1]), ": ", e$message)
  } else {
    errorMsg <- paste0("Error: ", e$message)
  }

  formattedStacktrace <- attr(tinytest, "formattedStacktrace")
  if (!is.na(formattedStacktrace)) {
    stacktraceDesc <- paste0("Stacktrace:\n\n", formattedStacktrace)
  } else {
    stacktraceDesc <- "No Stacktrace available!"
  }
  errorDescription <- paste(
    paste0("Uncaught error in test file: ", basename(attr(tinytest, "file"))),
    "No other tests from the test file could be captured duo the error.",
    errorMsg,
    "\n",
    stacktraceDesc,
    sep = "\n"
  )
  errorDescription <- escapeXmlText(errorDescription)
  errorTagMessage <- paste0(
    "Uncaught error in test file: ", basename(attr(tinytest, "file")), "\n", errorMsg
  )

  errorTag <- tag(
    name = "error",
    attributes = list(
      message = escapeXml(errorTagMessage),
      type = "Error"
    ),
    content = list(errorDescription)
  )
  testcaseTag <- tag(
    name = "testcase",
    attributes = list(
      classname = classnameTestcase(tinytest),
      name = nameTestcase(tinytest),
      status = "FAILED"
    ),
    content = list(errorTag)
  )
  return(testcaseTag)
  
}

#' Construct a testcase-tag for a side-effect test.
#' 
#' @param tinytest a `tinytest` object already validated to be a "SIDE-EFFECT" test.
#' @return a `testase` `XMLtag`
sideeffectTestcaseTag <- function(tinytest) {
  
  name <- "SIDE-EFFECT detected: "

  # Not sure if fst and/or lst will be defined for SIDE-EFFECT but in either case just 
  # add it anyway.
  fst <- attr(tinytest, "fst")
  lst <- attr(tinytest, "lst")
  if (!is.na(fst)) {
    name <- paste0(name, ": L", fst)
  }
  if (!is.na(lst) && (fst != lst)) {
    name <- paste0(name, "-L", lst)
  }
  testcaseTag <- tag(
    name = "testcase",
    attributes = list(
      classname = classnameTestcase(tinytest),
      name = escapeXml(name),
      status = "SIDE-EFFECT"
    ),
    content = list(
      tag(
        name = "system-out",
        content = list(
          constructFailureDescription(tinytest)
        )
      )
    )
  )
  return(testcaseTag)
}

#' XML tag
#' 
#' Create a list object that roughly mimics the behaviour of a simplistic XML tag element. 
#' Supported are XML tag-name, tag-attributes and tag-content.
#' 
#' @param name `character(1)` specifying the name of the tag.
#' @param attributes `named-list` being the XML attributes. 
#'    Names = attribute names, Values = attribute value.
#' @param content `unnamed-list` being the content XML-tag. Either child `XMLtag` or only  
#'  a single character vector being the xml text content. Currently no mixed-content is allowed.
#'  See details.
#'
#' 
#' @details
#' If a character vector is in the content it is converted to a single-length character vector. 
#' See [charVecToSingleLength()]
#'
#' Mixed content eg. a text string and a child xml tag next to each other is syntaxtically allows.
#' In practices it does not occur for XML that is schema formatted with XSD (like JUnit). 
#' So for simplicity it is not supported here.
#'
#' @return a `XMLtag`-object. 
tag <- function(name, attributes = list(), content = list()) {
  stopifnot(is.list(attributes), is.list(content), is.null(names(content)))
  if (length(attributes) > 0) {
    stopifnot(!is.null(names(attributes)), all(nchar(names(attributes)) > 0))
  }
  if (length(content) > 0) {
    contentIsString <- length(content) == 1 && is.character(content[[1]])
    contentIsChildTags <- all(vapply(content, inherits, "XMLtag", FUN.VALUE = logical(1)))
    if (!(contentIsString || contentIsChildTags)) {
      stop("Content should only be child-tags or only a single character vector")
    }
    if (contentIsString) content[[1]] <- charVecToSingleLength(content[[1]])
  }
  structure(list(name = name, attributes = attributes, content = content), class = "XMLtag")
}

#' Format method for XMLtag class
#' 
#' Format S3 method for the `XMLtag`-class
#' 
#' @param x an `XMLtag`-object
#' @param level print depth level. For each level 2 spaces are added to the left. The content of a 
#'   tag is automatically indented with 1 level. Except for text-content (see details).
#' 
#' @details
#' Note, text content does not get indented or put on a new line, since whites space characters 
#' are of relevance.
#' 
#' @param ... to ignore
#' @return `character(1)` vector of the formatted XML tag. 
#' @export
format.XMLtag <- function(x, level = 0, ...) {
  nSpacesTag <- strrep(" ", 2 * level)

  attributesStr <- vapply(
    X = names(x$attributes), 
    FUN = function(attrName) paste0(attrName, "='", x$attributes[[attrName]], "'"),
    FUN.VALUE = character(1L)
  )
  attributesStr <- paste0(attributesStr, collapse = " ")
  openingTag <- paste0("<", x$name, if (nchar(attributesStr) > 0) paste0(" ", attributesStr), ">")
  closingTag <- paste0("</", x$name, ">")
  
  if (length(x$content) == 0) {
    fTag <- paste0(nSpacesTag, openingTag, closingTag)
  } else if (length(x$content) == 1 && isSingleLengthCharNonNA(x$content[[1]])) {
    fTag <- paste0(nSpacesTag, openingTag, x$content[[1]], closingTag)   
  } else {
    fChildTags <- vapply(x$content, format, FUN.VALUE = character(1L), level = level + 1)
    fContent <- paste0(fChildTags, collapse = '\n')
    fTag <- paste0(nSpacesTag, openingTag, "\n", fContent, "\n", nSpacesTag, closingTag)
  }
  return(fTag)
}

#' Escape xml text 
#' 
#' Escape the characters '<' and `&` in a character vector meant to be xml-text content.
#' 
#' @param x a `character` vector meant to be xml-text content.
#' @return The same `character` vector x but xml text escaped.
escapeXmlText <- function(x) {
  stopifnot(is.character(x))
  x <- gsub("&", "&amp;", x)
  x <- gsub("<", "&lt;", x)
  return(x)
}


#' Escape xml
#' 
#' Escape the characters `&`,`"`,`'`,`<`,`>`
#' 
#' @seealso https://stackoverflow.com/a/1091953/10415129
#' @param x a `character` vector meant to be xml
#' @return The same `character` vector x but xml escaped.
escapeXml <- function(x) {
  stopifnot(is.character(x))
  x <- gsub("&", "&amp;", x)
  x <- gsub("<", "&lt;", x)
  x <- gsub(">", "&gt;", x)
  x <- gsub("'", "&apos;", x)
  x <- gsub('"', "&quot;", x)
  return(x)
}

#' Print method for XMLtag class.
#' 
#' Print method for XMLtag class.
#' 
#' @param x a `XMLtag`-object
#' @param ... to be ignored
#' @return `invisibly` the string that was printed to stdout.
#' @export
print.XMLtag <- function(x, ...) {
  str <- format(x)
  cat(str)
  invisible(str)
}

tag <- tinytest2JUnit:::tag

# tag is correctly constructed with an empty content
# Which is the case for an empty test folder!
resultEmpty <- tag(
  name = "testsuites",
  attributes = list(name = "tinytest results", tests = 0L, failures = 0L),
  content = list()
)

expect_inherits(resultEmpty, class = "XMLtag")
expect_equal(typeof(resultEmpty), target = "list")
expect_equal(resultEmpty$content, list())

# Format works correctly on an empty result
expect_inherits(format(resultEmpty), class = "character")
expect_true(length(format(resultEmpty)) == 1L)

expect_error(tag("t", content = list(0L)), info = "Only child-tags/char is in allowed content")
expect_error(
  tag("t", content = list(tag("t"), 0L)),
  info = "Only child-tags/char is in allowed content"
)
expect_error(
  tag("t", content = list(tag("t"), "text")),
  info = "Mixed tag content is currently not supported"
)
expect_error(
  tag("t", content = list("text", "text")),
  info = "Only a single character string is allowed."
)

# No errors are expected below:
tag("t", content = list("text"))
tag("t", content = list(tag("t")))
tag("t", content = list(tag("t"), tag('t2')))


expect_equal(
  format(tag("t", attributes = list(a = "1", b = "2"))),
  "<t a='1' b='2'></t>",
  info = "Empty content tags are closed on the same line."
)

expect_equal(
  format(tag("t", attributes = list(a = "1", b = "2"), content = list("Hello\nWorld"))),
  "<t a='1' b='2'>Hello\nWorld</t>",
  info = "White space is respected for text content"
)

formatTagWithChild <- format(
  tag("t", attributes = list(a = "1", b = "2"), content = list(tag("a"), tag("b")))
)
expectedFormattedTag <- "<t a='1' b='2'>\n  <a></a>\n  <b></b>\n</t>"
expect_equal(
  formatTagWithChild,
  expectedFormattedTag,
  info = "Child tags are indented and on a newline."
)

escaped <- tinytest2JUnit:::escapeXmlText(c("Hello <- world & !", "Leave > me ' \" alone"))
expect_equal(escaped, c("Hello &lt;- world &amp; !", "Leave > me ' \" alone"))


escaped <- tinytest2JUnit:::escapeXml(c("<a b='t' c=\"d\">&</a>", "&"))
expect_equal(escaped, c("&lt;a b=&apos;t&apos; c=&quot;d&quot;&gt;&amp;&lt;/a&gt;", "&amp;"))

# Project: tinytest2JUnit
# 
# Author: ltuijnder
###############################################################################

dir <- "~/git/tinytest2JUnit/inst/example_tests/simple_tests/"


add <- function(a, b){
  a + b
}

add_one <- function(a){
  a + 1
}

expect_equal(current = add(1,1), target = 2, info = "Math works.")
expect_equal(current = add_one(1), target = 2, info = "Math works with 1")
expect_equal(current = add_one(-1), target = 0)
expect_equal(current = add_one(-1), target = 1)

# Project: tinytest2JUnit
# 
# Author: ltuijnder
###############################################################################


#dir <- "~/git/tinytest2JUnit/inst/example_tests/heavy_calculations/"

heavyCalc <- function() {
  
  Sys.sleep(runif(1, 1, 3)) # sleep for 1 to 3 seconds
  TRUE
}


expect_true(heavyCalc(), info = "heavy calculations")

# Note This file is expected to be executed from tinytest/test_runTestDir.R
library(tinytest)

currentWD <- getwd()
filePath <- system.file(
  "example_tests/get_call_wd",
  package = "tinytest2JUnit"
)

callWD <- get_call_wd()

expect_equal(
  currentWD,
  filePath,
  info = "Tinytest executes test_files from within the directory of the test_file"
)

# Below assumes that this test_file executed from the tinytest/ directory!
# Also that devtools::load_all() is performed if you are testing interactively!
dirWhereItGotExecutedFrom <- system.file(
  "tinytest",
  package = "tinytest2JUnit"
)
expect_equal(
  callWD,
  dirWhereItGotExecutedFrom,
  info = "get_call_wd is correctly set."
)

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fibonacci.R
\name{fibonacci}
\alias{fibonacci}
\title{Fibonacci}
\usage{
fibonacci(n)
}
\arguments{
\item{n}{an integer}
}
\value{
A numeric vector of length n
}
\description{
Compute first n fibonacci numbers.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{charVecToSingleLength}
\alias{charVecToSingleLength}
\title{Convert any will character vector to a single length character vector}
\usage{
charVecToSingleLength(x)
}
\arguments{
\item{x}{a \code{character}}
}
\value{
x a single-length character vector Non-NA
}
\description{
Convert any will character vector to a single length character vector
}
\examples{
tinytest2JUnit:::charVecToSingleLength(c("Hello", "World")) # -> "HelloWorld"
tinytest2JUnit:::charVecToSingleLength(c("Hello", NA_character_)) # -> "HelloNA"
tinytest2JUnit:::charVecToSingleLength(character(0L)) # -> ""
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/writeJUnit.R
\name{writeJUnit}
\alias{writeJUnit}
\title{Write the results of a \code{tinytests}-object into JUnit xml report.}
\usage{
writeJUnit(tinytests, file = stdout(), overwrite = TRUE)
}
\arguments{
\item{tinytests}{\code{tinytests}-object to convert to JUnit xml.}

\item{file}{\code{character(1) | connection}: Full file path or connection object to write the
JUnit xml content to. By default \code{stdout()} connection is used.}

\item{overwrite}{\code{logical(1)}: should the file be overwritten if it already exist?
By default TRUE.}
}
\value{
\code{invisible(TRUE)} Might get another use in the future.
}
\description{
Write the \code{tinytests}-object to a JUnit XML reporting file. If a \code{tinytests2JUnit} is provided
(returned by \code{\link[=runTestDir]{runTestDir()}}) more info will get reported.
}
\section{Errors}{

In case of overwrite = FALSE and the file already exists an error is thrown.
}

\section{Side-effects}{

Side effects are registered as 'passed' tests in the JUnit output and have been given a status
"SIDE-EFFECT". The call and diff is also returned in the standard-output of the testcase tag.

They are not considred failures and would thus not stop a pipeline.
}

\section{tinytests to JUnit}{


To comply the the JUnit specification the tests results are adapted as follows:
\itemize{
\item A single test run \code{tinytests} is mapped to a \verb{<testsuites>} tag.
\item All \code{tinytest} results from a single file are mapped to a single \verb{<testsuite>} tag.
\itemize{
\item The name of the testsuite is equal to the test file name (without the file suffix)
}
\item An individual \code{tinytest} object (eg. a single \verb{except_*} exception test) is mapped to a
\verb{<testcase>} tag.
\itemize{
\item The name of the testcase is equal to the fileName + Line specification of where the expect
statement is performed + the info.
}
}

For reference: \url{https://github.com/testmoapp/junitxml}
}

\examples{
# Run tests with `tinytest`
dirWithTests <- system.file("example_tests/multiple_files",package = "tinytest2JUnit")
testresults <- runTestDir(dirWithTests)

writeJUnit(testresults) # Writes content to stdout

tmpFile <- tempfile(fileext = ".xml")
writeJUnit(tinytests = testresults, file = tmpFile)
}
\seealso{
The JUnit XML report format:  \url{https://github.com/testmoapp/junitxml}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/constructJUnitTag.R
\name{failureTestcaseTag}
\alias{failureTestcaseTag}
\title{Construct a testcase-tag for a failed test.}
\usage{
failureTestcaseTag(tinytest)
}
\arguments{
\item{tinytest}{a \code{tinytest} object already validated to be a "FAILURE" test.}
}
\value{
a \code{testase} \code{XMLtag}
}
\description{
Construct a testcase-tag for a failed test.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{isSingleLengthCharNonNA}
\alias{isSingleLengthCharNonNA}
\title{Test if single length character non NA.}
\usage{
isSingleLengthCharNonNA(x)
}
\arguments{
\item{x}{object to test.}
}
\value{
\code{logical(1)}
}
\description{
Test if single length character non NA.
}
\author{
ltuijnder
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/runTestDir.R
\name{runTestFile}
\alias{runTestFile}
\title{Internal wrapper arround tinytest::run_test_file}
\usage{
runTestFile(file, ...)
}
\arguments{
\item{file}{\code{character(1)} test file to run.}

\item{...}{arguments passed on to \code{\link[tinytest:run_test_file]{tinytest::run_test_file()}}}
}
\value{
a \code{tinytests2JUnit} object (being a subclass of \code{tinytest} object).
}
\description{
Internal wrapper arround \code{\link[tinytest:run_test_file]{tinytest::run_test_file()}} that records the test duration and
catches uncaught errors and logs the stacktrace of where the error occured.
}
\details{
The response is a subclass of the \code{tinytests} object called: \code{tinytests2Junit}
object which captures additional info for the reporting to JUnit:
\itemize{
\item Duration to run the file.
\item Timestamp when the test was run.
\item hostname of the computer where it was ran on.
}

The caught error is turned into a subclass uncaught-error of tinytest. This is implementation
detail and only to be understood by constructJUnitTag.

If an error occured it is captured and \code{uncaught-error} object (subclass of \code{tinytest}) is
returned in the \code{tinytests} object.
This tinytest object represents a "failed" tests that will get reported as an Error in the
JUnit. Various aspects of the error are also captured like the the stacktrace.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/constructJUnitTag.R
\name{constructTestcaseTag}
\alias{constructTestcaseTag}
\title{Construct JUnit \verb{</testcase>} tag}
\usage{
constructTestcaseTag(tinytest)
}
\arguments{
\item{tinytest}{a \code{tinytest}-object representing an individual test case.}
}
\value{
\code{XMLtag}: with tag-name = \code{tinytest} and contains the test result per test.
}
\description{
Construct JUnit \verb{</testcase>} tag based on a single \code{tinytest} result.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/runTestDir.R
\name{runTestDir}
\alias{runTestDir}
\title{Run all the test files in a directory}
\usage{
runTestDir(
  dir = "inst/tinytest",
  at_home = FALSE,
  pattern = "^test.*\\\\.[rR]$",
  cluster = NULL,
  lc_collate = getOption("tt.collate", NA),
  ...
)
}
\arguments{
\item{dir}{\code{character(1)} path to directory}

\item{at_home}{\code{logical(1)} should local tests be run? By default FALSE. Unlike
\code{\link[tinytest:run_test_dir]{tinytest::run_test_dir()}} which is meant to be called in a local interactive context. This
function is meant to be called in a non-interactive CI environment, where we want to mimic the
behaviour of how tests would get run by R CMD Check.
See also \code{at_home} documentation in tinytest package.}

\item{pattern}{\code{character(1)} A regular expression that is used to find
scripts in \code{dir} containing tests (by default \code{.R} or \code{.r}
files starting with \code{test}).}

\item{cluster}{A \code{cluster} object to run the test files on. Note, it is
expected that the clusters has already been prepared. Most notable, the package to test
should already been loaded. \code{runTestDir} will load the package
"tinytest" for you into the clusters. See \code{\link[tinytest:run_test_dir]{tinytest::run_test_dir()}} for more details.}

\item{lc_collate}{See \code{\link[tinytest:run_test_dir]{tinytest::run_test_dir()}}.}

\item{...}{Arguments passed on to \code{\link[tinytest:run_test_file]{tinytest::run_test_file()}}}
}
\value{
A \code{tinytests2Junit} object to be provided to the \code{\link[=writeJUnit]{writeJUnit()}} function.
}
\description{
\code{\link[=runTestDir]{runTestDir()}} is a drop in replacement for \code{\link[tinytest:run_test_dir]{tinytest::run_test_dir()}} with the key
difference that errors thrown from within a test file are caught and get reported with a
a stacktrace in the JUnit report. In addition, some extra metrics are recored for the JUnit
report, such as: timestamp, test duration, hostname and if tests are disabled (see details for
more info).
}
\details{
\code{\link[=runTestDir]{runTestDir()}} is meant as a CI-friendly alternative to the native \code{\link[tinytest:run_test_dir]{tinytest::run_test_dir()}}.
It catches errors that are raised in the tests files and adds them as a "failed" \code{tinytest}
in the output.

\code{\link[tinytest:run_test_dir]{tinytest::run_test_dir()}} would have let the error bubble up, stop the testing process and
not report any failures from other tests. One is then also forced to look into the
logs of the CI to see what the error was. The output of \code{\link[=runTestDir]{runTestDir()}} in combination with
\code{\link[=writeJUnit]{writeJUnit()}} will present you the error in the JUnit togheter with a stack trace. Next to the
test results of the other files that ran without a problem.

If you prefer the behaviour of \code{\link[tinytest:run_test_dir]{tinytest::run_test_dir()}} you can still use it in combination
\code{\link[=writeJUnit]{writeJUnit()}}.

Caught errors are returned in the output as as sub-class of \code{tinytest} object. This is however
considred implemenation detail and can be subject to change.

Note, function arguments explicilty listed in \code{\link[tinytest:run_test_dir]{tinytest::run_test_dir()}} but not here can still
still be provided via \code{...}
}
\section{tinytests2JUnit}{

The returned object is a \code{tinytests2JUnit} object (note the plural). This object
contains additional info compared to a \code{tinytests} object that is used in the JUnit report.

The following additional info will get reported:
\itemize{
\item The timestamp per test file on when it got invoked.
\item The test duration per test file.
\item The system hostname per test file on where it got invoked. This is mainly of interests for
different \code{clusters}.
\item If a test file is disabled. A test file is considered disabled if no tests
occur with in the file.
It is then assumed that at the top of file some conditional statement
made the test file exist early.
}
}

\examples{
# Run tests with `tinytest`
dirWithTests <- system.file("example_tests/multiple_files",package = "tinytest2JUnit")
testresults <- runTestDir(dirWithTests)

writeJUnit(testresults) # Writes content to stdout
}
\seealso{
\itemize{
\item \code{\link[tinytest:run_test_dir]{tinytest::run_test_dir()}} for how the function is inteded to behave.
\item \code{\link[=writeJUnit]{writeJUnit()}} where it is expected that the output of this function to be provided to.
\item \code{\link[=testPackage]{testPackage()}} for an higher-level function to simply test a package.
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/testPackage.R
\name{testPackage}
\alias{testPackage}
\title{Test an R package and report the results in JUnit}
\usage{
testPackage(
  pkgname,
  file = stdout(),
  errorOnFailure = TRUE,
  testdir = "tinytest",
  lib.loc = NULL,
  at_home = FALSE,
  ncpu = NULL,
  ...
)
}
\arguments{
\item{pkgname}{\code{character(1)}. Name of the package to tests.}

\item{file}{\code{character(1) | connection}: Full file path or connection object to write the
JUnit xml content to. By default \code{stdout()} connection is used.
\strong{Warning} if the file already exist it will be overwritten!}

\item{errorOnFailure}{\code{logical(1)} Should an error be raised (after writing the JUnit) when a
at least one test failed? By default TRUE. This is done as a
convenience to have the CI fail at the test stage on failure.}

\item{testdir}{\code{character(1)} testing directory of the package. See ?\code{\link[tinytest:test_package]{tinytest::test_package()}}
for more details.}

\item{lib.loc}{\code{character(1) | NULL} Library location where the package is installed.
By default: \code{NULL} meaning the package is searched on the standard .libPaths().}

\item{at_home}{\code{logical(1)} should local test be run? By default FALSE as we want to as closely
mimic the environement of how tests would get ran in CRAN.
See for more details \code{\link[tinytest:test_package]{tinytest::test_package()}}.}

\item{ncpu}{\verb{postive integer(1) | clutser} Either an integer specifying the amount of cpu's
to parralize the testing over or a \code{cluster} object to run the tests in.}

\item{...}{Extra arguments passed on to \code{\link[=runTestDir]{runTestDir()}}}
}
\value{
If \code{errorOnFailure} = FALSE, a \code{tinytests2JUnit} object (a subclass of \code{tinytests}
object that captures more info for export to JUnit). Else, an error is raised if at least
on failure occurs. Meant as convenience to automatically stop the CI build.
}
\description{
Run all tests of a package and report the results as JUnit xml. This function
can be seen as a drop in replacement for \code{\link[tinytest:test_package]{tinytest::test_package()}} but with a
key difference that uncaught errors will be catched and reported JUnit!
This function is intended to be used in a test stage of a CI build.
}
\details{
\code{\link[=testPackage]{testPackage()}} is meant as a CI-friendly alternative to the native \code{\link[tinytest:test_package]{tinytest::test_package()}}.
Next to directly reporting the tests results in a JUnit xml format, it also catches errors that
are raised in the tests files and reports them as "error" in the JUnit.

\code{\link[tinytest:test_package]{tinytest::test_package()}} would have let the error bubble up, stop the testing
process and not report any failures from other test files. One is then also forced
to look into the logs of the CI to see what the error was. \code{\link[=testPackage]{testPackage()}} presents you that
error in the JUnit with a stacktrace. Next to all the test results of the other files that
ran without a problem.

If you prefer the behaviour from \code{\link[tinytest:test_package]{tinytest::test_package()}}, you can still use it in
combination with \code{\link[=writeJUnit]{writeJUnit()}} if all tests results pass.

Just like \code{\link[tinytest:test_package]{tinytest::test_package()}} an error is raised if at least one failure occured during
testing. Obviously catched errors are also seen as failures. This error is raised
\strong{after} the test results have been written away to the file, such that your CI can still pick
it up and report the failure.
The error raising is done as a convenience to stop the CI from continue if test-failure occured.
You can opt-out of this behaviour by setting the \code{errorOnFailure} parameter to FALSE. Then a
case \code{tinytests2JUnit} object is returned (a sub-class of \code{tinytests} object containing addition
info for the JUnit).
Caught errors are also captured in this object as \code{tinytest}-objects. They
actually have a special sub-class but this is considered an internal implemenation detail.

\code{\link[=testPackage]{testPackage()}} is NOT meant to be called from within your \code{tests/tinytests.R} file! Tests
invoked by R CMD Check or on CRAN should still make use of \code{\link[tinytest:test_package]{tinytest::test_package()}}.
This function is only meant to be called from within a testing step in your CI to
report the test results in an JUnit xml format.
}
\section{Side-effects}{

Side effects are registered as 'passed' tests in the JUnit output and have been given a status
"SIDE-EFFECT". The call and diff is also returned in the standard-output of the testcase tag.

They are not considred failures and would thus not stop a pipeline.
}

\section{tinytests to JUnit}{


To comply the the JUnit specification the tests results are adapted as follows:
\itemize{
\item A single test run \code{tinytests} is mapped to a \verb{<testsuites>} tag.
\item All \code{tinytest} results from a single file are mapped to a single \verb{<testsuite>} tag.
\itemize{
\item The name of the testsuite is equal to the test file name (without the file suffix)
}
\item An individual \code{tinytest} object (eg. a single \verb{except_*} exception test) is mapped to a
\verb{<testcase>} tag.
\itemize{
\item The name of the testcase is equal to the fileName + Line specification of where the expect
statement is performed + the info.
}
}

For reference: \url{https://github.com/testmoapp/junitxml}
}

\examples{
tmpFile <- tempfile(fileext = ".xml")
testPackage("tinytest", file = tmpFile, verbose = 0)
}
\seealso{
\code{\link[=runTestDir]{runTestDir()}} and \code{\link[tinytest:test_package]{tinytest::test_package()}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/constructJUnitTag.R
\name{constructFailureDescription}
\alias{constructFailureDescription}
\title{Helper function generating the body of a failure description tag!}
\usage{
constructFailureDescription(tinytest)
}
\arguments{
\item{tinytest}{A \code{tinytest} objec that is considered failed!.}
}
\value{
\code{character(1)} being the failure tag description body. This string is already propery
xml escaped.
}
\description{
Helper function generating the body of a failure description tag! Attempts to mimic the print
behaviour of a tinytest object.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/xml.R
\name{print.XMLtag}
\alias{print.XMLtag}
\title{Print method for XMLtag class.}
\usage{
\method{print}{XMLtag}(x, ...)
}
\arguments{
\item{x}{a \code{XMLtag}-object}

\item{...}{to be ignored}
}
\value{
\code{invisibly} the string that was printed to stdout.
}
\description{
Print method for XMLtag class.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/constructJUnitTag.R
\name{constructTestsuiteTag}
\alias{constructTestsuiteTag}
\title{Construct JUnit \verb{</testsuite>} tag}
\usage{
constructTestsuiteTag(testsFile, id)
}
\arguments{
\item{testsFile}{\code{tinytests2JUnit | tinytests} -object with all test
results of a specified test file. At least a single test is expected.}

\item{id}{\code{integer(1)} testsuite id.}
}
\value{
\code{XMLtag}: with tag-name = \verb{</testsuite>}
that contains all the test results per test file.
}
\description{
Construct the \verb{</testsuite>} tag of a \code{tinytest}, given all the \code{tinytest} results
from a single test file.
}
\details{
In case a \code{tinytest2JUnit} is provided following additional info can be reported:
\itemize{
\item testsuite duration.
\item timestamp when the testsuite was performed.
\item hostname where the testsuite was ran.
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/constructJUnitTag.R
\name{nameTestcase}
\alias{nameTestcase}
\title{Helper function to construct the name of a testcase}
\usage{
nameTestcase(tinytest)
}
\arguments{
\item{tinytest}{a \code{tinytest} object. (does not matter what result)}
}
\value{
\code{character(1)} the testcase name to use for this tinytest object.
}
\description{
Helper function to construct the name of a testcase. Note, the charater is already xml escaped.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/xml.R
\name{escapeXmlText}
\alias{escapeXmlText}
\title{Escape xml text}
\usage{
escapeXmlText(x)
}
\arguments{
\item{x}{a \code{character} vector meant to be xml-text content.}
}
\value{
The same \code{character} vector x but xml text escaped.
}
\description{
Escape the characters '<' and \code{&} in a character vector meant to be xml-text content.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/xml.R
\name{format.XMLtag}
\alias{format.XMLtag}
\title{Format method for XMLtag class}
\usage{
\method{format}{XMLtag}(x, level = 0, ...)
}
\arguments{
\item{x}{an \code{XMLtag}-object}

\item{level}{print depth level. For each level 2 spaces are added to the left. The content of a
tag is automatically indented with 1 level. Except for text-content (see details).}

\item{...}{to ignore}
}
\value{
\code{character(1)} vector of the formatted XML tag.
}
\description{
Format S3 method for the \code{XMLtag}-class
}
\details{
Note, text content does not get indented or put on a new line, since whites space characters
are of relevance.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/constructJUnitTag.R
\name{classnameTestcase}
\alias{classnameTestcase}
\title{Helper function specifying the 'classname' attribute of the testcase tag.}
\usage{
classnameTestcase(tinytest)
}
\arguments{
\item{tinytest}{a \code{tinytest}-object representing an individual test case.}
}
\value{
\code{character(1)} being the 'classname'
}
\description{
Helper function specifying the 'classname' attribute of the testcase tag.
Currently equal to the fileName. The classname is already xml escaped.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/constructJUnitTag.R
\name{constructTestsuitesTag}
\alias{constructTestsuitesTag}
\title{Construct the JUnit \verb{</testsuites>} tag}
\usage{
constructTestsuitesTag(testResults)
}
\arguments{
\item{testResults}{\verb{tinytests2Junit | }tinytests\verb{object to convert into a JUnit XML object  Usually the result of [runTestDir()] for a}tinytests2JUnit\verb{object or native}tinytests`
produced by \code{\link[tinytest:test_package]{tinytest::test_package()}} or \code{\link[tinytest:run_test_dir]{tinytest::run_test_dir()}}.}
}
\value{
\code{XMLtag}: with tag-name = \verb{</testsuites>}. This is the root of the JUnit XML document.
}
\description{
Convert the \code{tinytests2Junit} or \code{tinytests}-object containing test across
possibly multiple files into a JUnit \verb{</testsuites>} tag. More details are reported to the
JUnit if a \code{tinytests2JUnit} object compared to the native \code{tinytests} object.
}
\details{
Reference for JUnit XML format:https://github.com/testmoapp/junitxml 

See details \code{\link[=runTestDir]{runTestDir()}} which additional info is recorded.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/runTestDir.R
\name{formattedFrame}
\alias{formattedFrame}
\title{Help function to generate the formatted string for a single stack frame.}
\usage{
formattedFrame(framecall, frameN, hasSrcInfo, dirName, fileName, lineNr)
}
\arguments{
\item{framecall}{\code{character(n)}: deparsed call for the given stack. Each element corresponds to
a line.}

\item{frameN}{\code{integer(1)}: frame nummer.}

\item{hasSrcInfo}{\code{logical(1)}: Does the call have any source info?}

\item{dirName}{\code{character(1)}: the directory name of the source file.}

\item{fileName}{\code{character(1)}: filename of the source. Value ignored if hasSrcInfo=TRUE.}

\item{lineNr}{\code{character(1)}: linenr in the source where the call occured..
Value ignored if hasSrcInfo=TRUE.}
}
\value{
\code{characer(1)} the formatted character string containing info of a single frame in the
stackstrace
}
\description{
Help function to generate the formatted string for a single stack frame.
}
\details{
For a given frame in the stack the string is formatted as follows (substitute the arguments
between the curly braces)
\code{
{frameN}| {call[1]}
{frameN}| {call[2]}
{frameN}| {call[3]}
{frameN}| {call[3]}
---> at File={dirName/fileName} Line={line}:
}

For example for only a single line error:
\code{
1: stop("This is a crash")
---> at File=R/my_r_code_file.R Line=234
}

Currently all call lines are printed for a given stack.
The last line with source file info only printed if hasSrcInfo=TRUE. Else it is ommited.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/xml.R
\name{tag}
\alias{tag}
\title{XML tag}
\usage{
tag(name, attributes = list(), content = list())
}
\arguments{
\item{name}{\code{character(1)} specifying the name of the tag.}

\item{attributes}{\code{named-list} being the XML attributes.
Names = attribute names, Values = attribute value.}

\item{content}{\code{unnamed-list} being the content XML-tag. Either child \code{XMLtag} or only
a single character vector being the xml text content. Currently no mixed-content is allowed.
See details.}
}
\value{
a \code{XMLtag}-object.
}
\description{
Create a list object that roughly mimics the behaviour of a simplistic XML tag element.
Supported are XML tag-name, tag-attributes and tag-content.
}
\details{
If a character vector is in the content it is converted to a single-length character vector.
See \code{\link[=charVecToSingleLength]{charVecToSingleLength()}}

Mixed content eg. a text string and a child xml tag next to each other is syntaxtically allows.
In practices it does not occur for XML that is schema formatted with XSD (like JUnit).
So for simplicity it is not supported here.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/runTestDir.R
\name{getFormattedStacktrace}
\alias{getFormattedStacktrace}
\title{Get formatted stack trace for an uncaught error from a tinytest test file.}
\usage{
getFormattedStacktrace()
}
\value{
\code{character(1)} a single length character string suitable to be printed to the end-user.
In case of no stakctrace (eg the error occured at root level of the script) NA_character_
is returned!
}
\description{
getFormattedStacktrace is a helper function that formats stacktrace for uncaught errors from a
tinytest run file.
}
\details{
This function assumes that it directly called from a withCallingHandler
error handling function! This fact is then used to remove the calling handler info from the
stack such that stack directly starts from where the error was thrown.

The function also removes the calls from the stack that involve
executing the \code{test_file}. The internals of runTestDir and tinytest are not of intererst.
And the highest level of the stack to consider is the top level of the test_file.

Note, this does mean that errors that occur on the top-level of the test file will not have a
a stacktrace!
For example: "Error: object 'x' not found" where x is attempted to be resolved
at the root levels
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/xml.R
\name{escapeXml}
\alias{escapeXml}
\title{Escape xml}
\usage{
escapeXml(x)
}
\arguments{
\item{x}{a \code{character} vector meant to be xml}
}
\value{
The same \code{character} vector x but xml escaped.
}
\description{
Escape the characters \code{&},\verb{"},\verb{'},\code{<},\code{>}
}
\seealso{
https://stackoverflow.com/a/1091953/10415129
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/constructJUnitTag.R
\name{passedTestcaseTag}
\alias{passedTestcaseTag}
\title{Construct a testcase-tag for a passed test.}
\usage{
passedTestcaseTag(tinytest)
}
\arguments{
\item{tinytest}{a \code{tinytest} object already validated to be a "PASSED" test.}
}
\value{
a \code{testase} \code{XMLtag}
}
\description{
Construct a testcase-tag for a passed test.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/runTestDir.R
\name{[.tinytests2JUnit}
\alias{[.tinytests2JUnit}
\alias{tinytests2JUnit}
\title{tinytestJUnit test results}
\usage{
\method{[}{tinytests2JUnit}(x, i)
}
\arguments{
\item{x}{\code{tinytests2JUnit} object to subset.}

\item{i}{object to subset the \code{x} with.}
}
\description{
An object of class \code{tinytests2JUnit}. Note the plurar. A subclass of \code{\link[tinytest:tinytests]{tinytest::tinytests()}}
containing extra info recordings that are used in the export to JUnit.
}
\details{
Following details are recorded when running the tests files and stored as additional attributes
to the object:
\itemize{
\item \strong{fileDurations}: \code{named-numeric(n)}. Names = filename of tests files, value = duration
in seconds on how long the test file took to run.
\item \strong{fileTimestamps}: \code{named-character(n)}. Names = filename of tests files, value = timestamp
when the test was invoked.
\item \strong{fileHostnames}: \code{named-character(n)}. Names = filename of tests files, value = The hostname
of the system that ran the tests. (Usefull in combination with clusters).
\item \strong{disabled}: \code{character}. A character vector of filenames where no tests were ran.
They are flagged as disabled tests.
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/constructJUnitTag.R
\name{errorTestcaseTag}
\alias{errorTestcaseTag}
\title{Construct a testcase-tag for an error test.}
\usage{
errorTestcaseTag(tinytest)
}
\arguments{
\item{tinytest}{a \code{tinytest} object already validated to be a "ERROR" test.}
}
\value{
a \code{testase} \code{XMLtag}
}
\description{
Construct a testcase-tag for an error test.
}
